module Labyrinth where

import VTree

-- Modelling labyrinths.

-- Our labyrinth is a Variadic trees (or a graph):
type Lab alpha = VTree alpha

-- A path in the tree is a list of node labels
type Path alpha = [alpha]

-- A leaf has no children
isLeaf :: Lab alpha-> Bool
isLeaf (Node a ns) = null ns

-- A leaf in an undirected graph has at most one child
isULeaf :: Lab alpha-> Bool
isULeaf (Node a ns) = null ns || null (tail ns)

-- All possible traversals, depth first via fold.
-- Does not work for graphs (infinite trees), as it will diverge!
dfts' :: Lab alpha-> [Path alpha]
dfts' = foldT add where
  add a [] = [[a]]
  add a ps = concatMap (map (a :)) ps

-- All possible traversals, depth first
dfts :: Eq alpha=> (Lab alpha-> Bool)-> Lab alpha-> [Path alpha]
dfts trg = dfts0 [] where
  dfts0 p n@(Node a ns)
    | trg n     = [reverse (a:p)]
    | elem a p  = [] 
    | otherwise = concatMap (dfts0 (a:p)) ns 

-- All possible traversals, breadth first
bfts :: Lab alpha-> [Path alpha]
bfts l = bfts0 [] [l] where
  bfts0 p [] = []
  bfts0 p (Node a cs:ns) =
    reverse (a:p) : (bfts0 p ns ++ bfts0 (a:p) cs)

-- A path contains a cycle if the same node appears twice
cycles :: Eq alpha=> Path alpha-> Bool
cycles [] = False
cycles (a:ps) = any (a ==) ps || cycles ps 


-- An identifier for the nodes:
type Id  = Int

-- An example labyrinth without cycles:
l00 = Node 0 [l01]
l10 = Node 1 [l00]
l20 = Node 2 [l21]
l30 = Node 3 [l20]
l40 = Node 4 []
l01 = Node 5 [l02 ]
l11 = Node 6 []
l21 = Node 7 [l11]
l31 = Node 8 [l30, l41]
l41 = Node 9 [l40, l42]
l02 = Node 10 [l12, l03]
l12 = Node 11 [l13]
l22 = Node 12 [l32]
l32 = Node 13 [l31]
l42 = Node 14 []
l03 = Node 15 [l04]
l13 = Node 16 [l23]
l23 = Node 17 [l22, l24 ]
-- l23 = Node 17 [l24 -- this change bisects the labyrinth]
l33 = Node 18 [l43]
l43 = Node 19 [l44]
l04 = Node 20 [l14]
l14 = Node 21 []
l24 = Node 22 [l34]
l34 = Node 23 [l33]
l44 = Node 24 []


-- An example labyrinth with cycles:
m00 = Node 0 [m01]
m10 = Node 1 [m00]
m20 = Node 2 [m21]
m30 = Node 3 [m20]
m40 = Node 4 []
m01 = Node 5 [m02] -- Alternatively, Node 5 [m11, m02]
m11 = Node 6 [m01] -- This makes one cycle
m21 = Node 7 [m11]
m31 = Node 8 [m30, m41]
m41 = Node 9 [m40, m42]
m02 = Node 10 [m12, m03]
m12 = Node 11 [m13]
m22 = Node 12 [m32]
m32 = Node 13 [m31]
m42 = Node 14 []
m03 = Node 15 [m04]
m13 = Node 16 [m23]
m23 = Node 17 [m22, m24]
m33 = Node 18 [m43]
m43 = Node 19 [m44]
m04 = Node 20 [m14]
m14 = Node 21 [m24] -- This makes the other cycle
m24 = Node 22 [m34] -- Alternatively, Node 22 [m14, m34]
m34 = Node 23 [m33]
m44 = Node 24 []

-- The second labyrinth, with undirected edges
u0  = Node 0 [u1, u5]
u1  = Node 1 [u0]
u2  = Node 2 [u3, u7]
u3  = Node 3 [u2, u8]
u4  = Node 4 [u9]
u5  = Node 5 [u0, u6, u10]
u6  = Node 6 [u5, u7]
u7  = Node 7 [u2, u6]
u8  = Node 8 [u3, u9, u13]
u9  = Node 9 [u4, u8, u14]
u10 = Node 10 [u5, u11, u15]
u11 = Node 11 [u10, u16]
u12 = Node 12 [u13, u17]
u13 = Node 13 [u12, u8]
u14 = Node 14 [u9]
u15 = Node 15 [u10, u20]
u16 = Node 16 [u11, u17]
u17 = Node 17 [u12, u16, u22]
u18 = Node 18 [u19, u23]
u19 = Node 19 [u18, u24]
u20 = Node 20 [u15, u21]
u21 = Node 21 [u20, u22]
u22 = Node 22 [u17, u21, u23]
u23 = Node 23 [u18, u22]
u24 = Node 24 [u19]
